"""pyTractoViz package.

Python tools for diffusion tractography visualization
"""

from __future__ import annotations

from pytractoviz._internal.cli import get_parser, main
from pytractoviz.html import create_quality_check_html

__all__: list[str] = ["create_quality_check_html", "get_parser", "main"]
